#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>

/* The part of the file before main() implements a persistent AVL
   tree that maps string keys to tree values.

   An AVL tree is a balanced binary tree, so that keys can be found
   quickly.

   A persistent tree is one where adding to the tree is a functional
   operation: the result is a new tree with the additional mapping,
   and the old tree remains as-is.

   A persistent AVL tree is interesting, because inserting into the
   tree alocates new nodes along a spine of a modified tree, and the
   new and old trees otherwise share subtress. Rotations to preserve
   balance make the pattern of sharing complex.

   This first implementation calls malloc() without free(), so it uses
   too much memory.

   The main() function applies AVL trees to read in a sequence of
   words and detect whether a given word appears before another given
   word in the document. Persistent AVL trees make that search fast
   (due to being a balanced binary tree) and memory-efficient (because
   each word can be mapped to a tree that shares structure with
   other word mappings).

   We allocate strings in a string() record to set up for better
   memory management. */

typedef struct string {
  char str[0];
} string;

string *make_string(char *s) {
  size_t len = strlen(s) + 1;
  string *key = malloc(sizeof(string) + len);
  memcpy(key->str, s, len);
  return key;
}

typedef struct node tree;

struct node {
  string *key;
  tree *val;
  int height;
  struct node *left;  
  struct node *right;
};

int get_height(tree *t) {
  if (t == NULL)
    return 0;
  else
    return t->height;
}

void fix_height(tree *t) {
  int h = get_height(t->left);
  if (get_height(t->right) > h)
    h = get_height(t->right);
  t->height = h + 1;
}

tree *leaf(string *key, tree *val) {
  tree *t = malloc(sizeof(struct node));

  t->key = key;
  t->val = val;
  t->left = NULL;
  t->right = NULL;
  t->height = 1;

  return t;
}

tree *branch(string *key, tree *val, tree *l, tree *r) {
  tree *t = malloc(sizeof(struct node));

  t->key = key;
  t->val = val;
  t->left = l;
  t->right = r;
  fix_height(t);

  return t;
}

tree *clone_branch(tree *t) {
  return branch(t->key, t->val, t->left, t->right);
}

int get_balance(tree *t) {
  return get_height(t->left) - get_height(t->right);
}

/****************************************/

tree *btsearch(tree *t, string *key)
{
  while (t != NULL) {
    int cmp = strcmp(t->key->str, key->str);
    if (cmp > 0)
      t = t->left;
    else if (cmp < 0)
      t = t->right;
    else
      return t->val;
  }
  return NULL;
}

/****************************************/

tree *btinsert_left(tree *t, string *key, tree *val);
tree *btinsert_right(tree *t, string *key, tree *val);

tree *btinsert(tree *t, string *key, tree *val)
{
  if (t == NULL)
    return leaf(key, val);
  else {
    int cmp = strcmp(t->key->str, key->str);
    if (cmp > 0)
      return btinsert_left(t, key, val);
    else if (cmp < 0)
      return btinsert_right(t, key, val);
    else
      return t;
  }
}

tree *btinsert_left(tree *t, string *key, tree *val) {
  /* insert smaller number to left */
  t = clone_branch(t);
  t->left = btinsert(t->left, key, val);
  fix_height(t);

  if (get_balance(t) == 2) {
    /* need to rotate right */
    tree *left = clone_branch(t->left);
    if (get_balance(left) < 0) {
      /* double right rotation */
      tree *left_right = clone_branch(left->right);
      left->right = left_right->left;
      left_right->left = left;
      fix_height(left);
      left = left_right;
    }
    t->left = left->right;
    left->right = t;
    fix_height(t);
    fix_height(left);
    return left;
  }
   
  return t;
}

tree *btinsert_right(tree *t, string *key, tree *val) {
  /* insert larger number to right */
  t = clone_branch(t);
  t->right = btinsert(t->right, key, val);
  fix_height(t);

  if (get_balance(t) == -2) {
    /* need to rotate left */
    tree *right = clone_branch(t->right);
    if (get_balance(right) > 0) {
      /* double left rotation */
      tree *right_left = clone_branch(right->left);
      right->left = right_left->right;
      right_left->right = right;
      fix_height(right);
      right = right_left;
    }
    t->right = right->left;
    right->left = t;
    fix_height(t);
    fix_height(right);
    return right;
  }
  
  return t;
}

/****************************************/

long get_memory_use() {
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return ru.ru_maxrss;
}

/****************************************/

int main(int argc, char** argv)
{
  tree *t = NULL, *prev_t;
  string *key;
  char buffer[256];
  FILE *f;

  if (argc != 2) {
    fprintf(stderr, "usage: %s <file>\n", argv[0]);
    return 1;
  }

  /* read tree from a file and don't print */
  
  f = fopen(argv[1], "r");
  if (!f) {
    fprintf(stderr, "could not open %s\n", argv[1]);
    return 1;
  }
  while (fgets(buffer, 256, f)) {
    t = btinsert(t, make_string(buffer), t);
  }
  fclose(f);

  printf("Memory use is %ld\n", get_memory_use());
  
  printf("Ready to seach\n");
  while (fgets(buffer, 256, stdin)) {
    key = make_string(buffer);
    prev_t = btsearch(t, key);
    if (prev_t == NULL)
      printf("... not there or nothing before\n");
    else {
      fgets(buffer, 256, stdin);
      key = make_string(buffer);
      if (btsearch(prev_t, key))
        printf("... yes, appeared before\n");
      else
        printf("... no, did not appear before\n");
    }
  }

  return 0;
}
